import time
import functools
from typing import Callable, Literal, TypeVar, Any

from .global_logger import get_global_log

F = TypeVar("F", bound=Callable[..., Any])


class Profiler:
    def __init__(self, unit: Literal["ns", "ms", "s"] = "s") -> None:
        self.module: str = ""
        self.name: str = ""
        self.unit: Literal["ns", "ms", "s"] = unit
        self.factor: float = {
            "ns": 1.0,
            "ms": 1e-6,
            "s": 1e-9,
        }[unit]
        self.times_called: int = 0
        self.elapsed_time_total: float = 0
        self.elapsed_time_max: float = 0
        self.cpu_time_total: float = 0
        self.cpu_time_max: float = 0

    def __repr__(self):
        msg = (
            f"Function: {self.module}.{self.name}\n"
            f"Times called: {self.times_called}\n"
            f"Elapsed Time (Total: {self.elapsed_time_total} {self.unit}, Avg: {self.elapsed_time_avg} {self.unit}, Max: {self.elapsed_time_max} {self.unit})\n"
            f"CPU Time (Total: {self.cpu_time_total} {self.unit}, Avg: {self.cpu_time_avg} {self.unit}, Max: {self.cpu_time_max} {self.unit}, Ratio: {self.cpu_time_ratio})\n"
        )
        return msg

    def __call__(self, func: F) -> F:
        self.module = func.__module__
        self.name = func.__name__

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            # if the global log is set, store this Profiler instance
            log = get_global_log()
            if log is not None:
                log.insert_profiler(f"{self.module}.{self.name}", self)

            start_elapsed_time = time.perf_counter_ns()
            start_cpu_time = time.process_time_ns()
            retval = func(*args, **kwargs)
            end_elapsed_time = time.perf_counter_ns()
            end_cpu_time = time.process_time_ns()

            self.times_called += 1

            it_elapsed_time = round((end_elapsed_time - start_elapsed_time) * self.factor, 9)
            it_cpu_time = round((end_cpu_time - start_cpu_time) * self.factor, 9)
            self.elapsed_time_total = round(self.elapsed_time_total + it_elapsed_time, 9)
            self.cpu_time_total = round(self.cpu_time_total + it_cpu_time, 9)
            self.elapsed_time_max = max(self.elapsed_time_max, it_elapsed_time)
            self.cpu_time_max = max(self.cpu_time_max, it_cpu_time)
            if self.cpu_time_max > 1:
                pass
            return retval

        wrapper.__setattr__("profiler", self)
        return wrapper  # pyright: ignore[reportReturnType]

    @property
    def elapsed_time_avg(self) -> float:
        return (
            round(self.elapsed_time_total / self.times_called, 9) if self.times_called > 0 else 0.0
        )

    @property
    def cpu_time_avg(self) -> float:
        return round(self.cpu_time_total / self.times_called, 9) if self.times_called > 0 else 0.0

    @property
    def cpu_time_ratio(self) -> float:
        return (
            round(self.cpu_time_total / self.elapsed_time_total, 9)
            if self.times_called > 0
            else 0.0
        )



import os
import shutil
from datetime import datetime
from typing import Any
from threading import Lock

from .config import image_enabled, web_enabled, screenshot, Image


class ScreenshotManager:
    """Manages screenshot capture functionality for automation logging."""

    def __init__(self, log_dir: str, logger: Any):
        """
        Initialize screenshot manager.

        Parameters
        ----------
        log_dir : str
            Directory where screenshots will be saved
        logger : Any
            Logger instance for logging screenshot events
        """
        self.log_dir: str = log_dir
        self._logger = logger
        self._mutex: Lock = Lock()

    def capture_screenshot(self, filename: str | None = None, optimize_size: bool = False) -> str:
        """
        Captures a screenshot of the entire screen and saves it in the log directory.

        The name of the screenshot is written with level INFO in the log file with the format:
        screenshot_log_timestamp.png (default) or the filename parameter.

        Parameters
        ----------
        filename : str, optional
            Name of the screenshot. If no extension is found in filename, it will be .png if `optimize_size` = False.
            If `optimize_size` is True, the extension will be .webp whether the extension is found or not.

            If a file with the same name already exists, the new file will be named with a suffix (x), where x is
            the number of existing files with the same name. Example: image.png, image (1).png,
            image(2).png, image(3).png.

        optimize_size : bool
            Flag to save the image using a memory efficient and lossless format (.webp)

        Returns
        -------
        filename : str
            The name of the file saved

        Raises
        ------
        NotImplementedError
            If pyautogui and pillow/PIL are not installed
        ValueError
            If the `filename` passed is not a string

        Examples
        --------
        >>> log = AutomationLogger(script_path=__file__)
        >>> log.capture_screenshot("example_screenshot.png")
        """

        if not image_enabled:
            raise NotImplementedError(
                "Function is only available if pyautogui and pillow/PIL are installed"
            )

        if filename is not None and not isinstance(filename, str):
            raise ValueError("filename must be a string")

        if filename is None:
            timestamp = datetime.now().strftime(r"%Y_%m_%d-%H_%M_%S")
            basename = f"screenshot_{timestamp}"
            extension = ".png"
        else:
            basename, extension = os.path.splitext(filename)
            if extension == "":
                extension = ".png"

        if optimize_size:
            extension = ".webp"

        filename = basename + extension

        with self._mutex:
            suffix = 0
            # If file with same name exists, include a suffix (number)
            while os.path.exists(os.path.join(self.log_dir, filename)):
                suffix += 1
                filename = f"{basename} ({suffix}){extension}"
            filepath = os.path.join(self.log_dir, filename)
            img: Image = screenshot(filepath)
            if optimize_size:
                img.save(filepath, lossless=True, optimize=True)
            self._logger.info(f"Screenshot captured and saved as: {filename}")

        return filename

    def capture_screenshot_selenium(self, driver: Any, filename: str | None = None) -> str:
        """
        Captures a screenshot of a Selenium driver instance.

        The name of the screenshot is written with level INFO in the log file with the format:
        selenium_screenshot_log_timestamp.png or the filename parameter.

        Parameters
        ----------
        driver : WebDriver
            Selenium WebDriver instance from which to capture the screenshot.

        filename : str, optional
            Name of the screenshot. If a file with the same name already exists, the new file
            will be named with a suffix (x), where x is the number of existing files with the
            same name. Example: image.png, image (1).png, image(2).png, image(3).png.

        Returns
        -------
        filename : str
            The name of the file saved

        Raises
        ------
        NotImplementedError
            If selenium is not installed
        ValueError
            If the `filename` passed is not a string

        Examples
        --------
        >>> log = AutomationLogger(script_path=__file__)
        >>> driver = selenium.webdriver.Chrome(...)
        >>> log.capture_screenshot_selenium(driver, "chromedriver_screenshot.png")
        """

        if not web_enabled:
            raise NotImplementedError("Function is only available if selenium is installed")

        if filename is not None and not isinstance(filename, str):
            raise ValueError("filename must be a string")

        if filename is None:
            timestamp = datetime.now().strftime(r"%Y_%m_%d-%H_%M_%S")
            basename = f"selenium_screenshot_{timestamp}"
        else:
            basename = os.path.splitext(filename)[0]
        filename = basename + ".png"

        with self._mutex:
            # If file with same name exists, include a suffix (number)
            suffix = 0
            while os.path.exists(os.path.join(self.log_dir, filename)):
                suffix += 1
                filename = f"{basename} ({suffix}).png"

            driver.save_screenshot(os.path.join(self.log_dir, filename))
            self._logger.info(f"Selenium screenshot captured and saved as: {filename}")

        return filename

    def group_by_prefix(self, prefix: str | None = None, sep: str | None = None):
        """
        Group files in the log directory by prefix.
        If `prefix` is given all files that don't have the prefix will be ignored.
        If `sep` is given all files will be grouped according to automatic prefixes defined by the
        text before the separator.
        `prefix` takes priority over `sep` if both are set

        This function is useful to group images captured during the execution

        Parameters
        ----------
        prefix: str, optional
            To ensure expected behaviour, the prefix in the file name must be followed by one of the following: [" ", "-", "_", "@"]

        sep: str, optional
            Separator used to split file names and automatically determine file prefixes.
            To ensure expected behaviour, the string before the separator must be followed by one of the following: [" ", "-", "_", "@", sep]

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If the `prefix` and `sep` are None

        Examples
        --------
        >>> log = AutomationLogger(script_path=__file__)
        >>> # Manual prefix
        >>> log.capture_screenshot("ABC - Image 1")
        >>> log.capture_screenshot("ABC - Image 2")
        >>> log.capture_screenshot("ABC - Image 3")
        >>> log.group_by_prefix(prefix="ABC") # creates folder "ABC"
        >>> # Automatic Prefix
        >>> log.capture_screenshot("ABC - Image 1")
        >>> log.capture_screenshot("ABC - Image 2")
        >>> log.capture_screenshot("ABCD - Image 1")
        >>> log.group_by_prefix(sep="-") # creates folders "ABC" and "ABCD"
        """

        if prefix is None and sep is None:
            raise ValueError("prefix or sep must be set")

        with self._mutex:
            if prefix is not None:
                prefixes = [str(prefix)]
            else:
                sep = str(sep)
                files = [x for x in os.listdir(self.log_dir) if sep in x and x != self._logger.name]
                prefixes = set(x.split(sep)[0] for x in files)

            # There was a bug when a prefix is a substring of another, such as prefix1 and prefix10
            # An attempt to avoid it is to see if the prefix is separate from the rest of the text
            # by common separators
            separators = [" ", "_", "-", "@"]
            if sep is not None:
                separators.append(sep)

            for prefix in prefixes:
                prefix = prefix.strip()
                files: list[str] = []
                for file in os.listdir(self.log_dir):
                    if file == self._logger.name or not os.path.isfile(
                        os.path.join(self.log_dir, file)
                    ):
                        continue
                    if prefix[-1] in separators:
                        files.append(prefix[-1])
                    else:
                        for sep in separators:
                            if file.startswith(prefix + sep):
                                files.append(file)
                                break

                os.makedirs(os.path.join(self.log_dir, prefix), exist_ok=True)
                for file in files:
                    _ = shutil.move(
                        os.path.join(self.log_dir, file),
                        os.path.join(self.log_dir, prefix),
                    )
                self._logger.info(f"Grouped files by prefix: {prefix}")



from typing import Any

from .protocols import IAutomationLogger

# Global logger instance
global_log: IAutomationLogger | None = None


def get_global_log() -> IAutomationLogger | None:
    """
    Returns the value of the global_log variable
    """

    return global_log


def set_global_log(logger: IAutomationLogger) -> None:
    """
    Set the global log object of the `automation_logging` module

    Parameters
    ----------
    logger: AutomationLogger.
        The instance of AutomationLogger that will be used globally

    Returns
    -------
    None

    Examples
    --------
    >>> log_dir = "./tests/Logs"
    >>> log = AutomationLogger(script_path=__file__, log_dir=log_dir, log_to_console=True)
    >>> set_global_log(log)
    >>> info("This was printed by the global log")

    Notes
    -----
    This function is thread-safe
    """

    global global_log
    global_log = logger


def debug(message: str) -> None:
    """Writes the message to the log file with level DEBUG

    Parameters
    ----------
    message: str
        Message to be logged

    Returns
    -------
    None

    Raises
    ------
    ValueError
        If global log is not set

    Examples
    --------
    >>> log = AutomationLogger(script_path=__file__)
    >>> set_global_log(log)
    >>> debug(f"This is a DEBUG example")

    Notes
    -----
    This function is thread-safe
    """
    if global_log is None:
        raise ValueError("This function requires the global log to be set")
    global_log.debug(message)


def debug_else(message: str) -> None:
    """Calls debug if global_log is set, else calls print"""
    if global_log is not None:
        debug(message)
    else:
        print(f"{'DEBUG':<10} | {message}")


def info(message: str) -> None:
    """Writes the message to the log file with level INFO

    Parameters
    ----------
    message: str
        Message to be logged

    Returns
    -------
    None

    Raises
    ------
    ValueError
        If global log is not set

    Examples
    --------
    >>> log = AutomationLogger(script_path=__file__)
    >>> set_global_log(log)
    >>> info(f"This is an INFO example")

    Notes
    -----
    This function is thread-safe
    """
    if global_log is None:
        raise ValueError("This function requires the global log to be set")
    global_log.info(message)


def info_else(message: str) -> None:
    """Calls info if global_log is set, else calls print"""
    if global_log is not None:
        info(message)
    else:
        print(f"{'INFO':<10} | {message}")


def stat(message: str | dict[Any, Any]) -> None:
    """Writes the message to the log file with level STAT

    Parameters
    ----------
    message: str | dict[Any, Any]
        Message to be logged

    Returns
    -------
    None

    Raises
    ------
    ValueError
        If global log is not set

    Examples
    --------
    >>> log = AutomationLogger(script_path=__file__)
    >>> set_global_log(log)
    >>> info_dict = {"program_name": "Test Program", "var1": "StringVariable", "var2": 3}
    >>> stat(info_dict)

    Notes
    -----
    This function is thread-safe
    """
    if global_log is None:
        raise ValueError("This function requires the global log to be set")
    global_log.stat(message)


def stat_else(message: str) -> None:
    """Calls stat if global_log is set, else calls print"""
    if global_log is not None:
        stat(message)
    else:
        print(f"{'STAT':<10} | {message}")


def warning(message: str) -> None:
    """Writes the message to the log file with level WARNING

    Parameters
    ----------
    message: str
        Message to be logged

    Returns
    -------
    None

    Raises
    ------
    ValueError
        If global log is not set

    Examples
    --------
    >>> log = AutomationLogger(script_path=__file__)
    >>> set_global_log(log)
    >>> warning(f"This is a WARNING example")

    Notes
    -----
    This function is thread-safe
    """
    if global_log is None:
        raise ValueError("This function requires the global log to be set")
    global_log.warning(message)


def warning_else(message: str) -> None:
    """Calls warning if global_log is set, else calls print"""
    if global_log is not None:
        warning(message)
    else:
        print(f"{'WARNING':<10} | {message}")


def error(message: str) -> None:
    """Writes the message to the log file with level ERROR

    Parameters
    ----------
    message: str
        Message to be logged

    Returns
    -------
    None

    Raises
    ------
    ValueError
        If global log is not set

    Examples
    --------
    >>> log = AutomationLogger(script_path=__file__)
    >>> set_global_log(log)
    >>> error(f"This is an ERROR example")

    Notes
    -----
    This function is thread-safe
    """
    if global_log is None:
        raise ValueError("This function requires the global log to be set")
    global_log.error(message)


def error_else(message: str) -> None:
    """Calls error if global_log is set, else calls print"""
    if global_log is not None:
        error(message)
    else:
        print(f"{'ERROR':<10} | {message}")


def exception(message: str) -> None:
    """Writes the message to the log file with level EXCEPTION

    Parameters
    ----------
    message: str
        Message to be logged

    Returns
    -------
    None

    Raises
    ------
    ValueError
        If global log is not set

    Examples
    --------
    >>> log = AutomationLogger(script_path=__file__)
    >>> set_global_log(log)
    >>> try:
    >>>     a = 1/0
    >>> except Exception as exc:
    >>>     exception(f"This is an EXCEPTION example: {repr(exc)}")

    Notes
    -----
    This function is thread-safe
    """
    if global_log is None:
        raise ValueError("This function requires the global log to be set")
    global_log.exception(message)


def exception_else(message: str) -> None:
    """Calls exception if global_log is set, else calls print"""
    if global_log is not None:
        exception(message)
    else:
        print(f"{'EXCEPTION':<10} | {message}")


def critical(message: str) -> None:
    """Writes the message to the log file with level CRITICAL

    Parameters
    ----------
    message: str
        Message to be logged

    Returns
    -------
    None

    Raises
    ------
    ValueError
        If global log is not set

    Examples
    --------
    >>> log = AutomationLogger(script_path=__file__)
    >>> set_global_log(log)
    >>> is_valid = check_critical_component(...)
    >>> if not is_valid:
    >>>     critical("This is a CRITICAL example, cannot continue")
    >>>     sys.exit(2)

    Notes
    -----
    This function is thread-safe
    """
    if global_log is None:
        raise ValueError("This function requires the global log to be set")
    global_log.critical(message)


def critical_else(message: str) -> None:
    """Calls critical if global_log is set, else calls print"""
    if global_log is not None:
        critical(message)
    else:
        print(f"{'CRITICAL':<10} | {message}")


def capture_screenshot(filename: str | None = None, optimize_size: bool = False) -> str:
    """
    Captures a screenshot of the entire screen and saves it in the log directory.

    The name of the screenshot is written with level INFO in the log file with the format:
    screenshot_log_timestamp.png (default) or the filename parameter.

    Parameters
    ----------
    filename : str, optional
        Name of the screenshot. If no extension is found in filename, it will be .png if `optimize_size` = False.
        If `optimize_size` is True, the extension will be .webp whether the extesion is found or not.

        If a file with the same name already exists, the new file will be named with a suffix (x), where x is
        the number of existing files with the same name. Example: image.png, image (1).png,
        image(2).png, image(3).png.

    optimize_size : bool
        Flag to save the image using a memory efficient and lossless format (.webp)

    Returns
    -------
    filename : str
        The name of the file saved

    Raises
    ------
    ValueError
        If global log is not set

    Examples
    --------
    >>> log_dir = "./tests/Logs"
    >>> log = AutomationLogger(script_path=__file__, log_dir=log_dir, log_to_console=True)
    >>> set_global_log(log)
    >>> capture_screenshot("example_screenshot.png")

    Notes
    -----
    This function is thread-safe
    """
    if global_log is None:
        raise ValueError("This function requires the global log to be set")
    return global_log.capture_screenshot(filename, optimize_size)


def capture_screenshot_selenium(driver: Any, filename: str | None = None) -> str:
    """
    Captures a screenshot of a Selenium driver instance.

    The name of the screenshot is written with level INFO in the log file with the format:
    log_timestamp_screenshot_selenium.png or the filename parameter.

    Parameters
    ----------
    driver : WebDriver
        Selenium WebDriver instance from which to capture the screenshot.

    filename : str, optional
        Name of the screenshot. If a file with the same name already exists, the new file
        will be named with a suffix (x), where x is the number of existing files with the
        same name. Example: image.png, image (1).png, image(2).png, image(3).png.

    Returns
    -------
    filename : str
        The name of the file saved

    Raises
    ------
    ValueError
        If global log is not set

    Examples
    --------
    >>> log_dir = "./tests/Logs"
    >>> log = AutomationLogger(script_path=__file__, log_dir=log_dir, log_to_console=True)
    >>> set_global_log(log)
    >>> driver = selenium.webdriver.Chrome(...)
    >>> capture_screenshot_selenium(driver, "chromedriver_screenshot.png")

    Notes
    -----
    This function is thread-safe
    """
    if global_log is None:
        raise ValueError("This function requires the global log to be set")
    return global_log.capture_screenshot_selenium(driver, filename)


def group_by_prefix(prefix: str | None = None, sep: str | None = None) -> None:
    """
    Group files in the log directory by prefix.
    If `prefix` is given all files that don't have the prefix will be ignored.
    If `sep` is given all files will be grouped according to automatic prefixes defined by the
    text before the separator.
    `prefix` takes priority over `sep` if both are set

    This function is useful to group images captured during the execution

    Parameters
    ----------
    prefix: str, optional
        Prefix of files that will be grouped inside the folder `prefix` in the log directory

    sep: str, optional
        Separator used to split file names and automatically determine file prefixes.
        Files with the same prefix will be grouped inside the same folder in the log directory

    Returns
    -------
    None

    Raises
    ------
    ValueError
        If global log is not set

    Examples
    --------
    >>> log_dir = "./tests/Logs"
    >>> log = AutomationLogger(script_path=__file__, log_dir=log_dir, log_to_console=True)
    >>> set_global_log(log)
    >>> # Manual prefix
    >>> capture_screenshot("ABC - Image 1")
    >>> capture_screenshot("ABC - Image 2")
    >>> capture_screenshot("ABC - Image 3")
    >>> group_by_prefix(prefix="ABC") # creates folder "ABC"
    >>> # Automatic Prefix
    >>> capture_screenshot("ABC - Image 1")
    >>> capture_screenshot("ABC - Image 2")
    >>> capture_screenshot("ABCD - Image 1")
    >>> group_by_prefix(sep="-") # creates folders "ABC" and "ABCD"

    Notes
    -----
    This function is thread-safe
    """
    if global_log is None:
        raise ValueError("This function requires the global log to be set")
    global_log.group_by_prefix(prefix, sep)


def log_profilers() -> str:
    """
    Write profilers captured by the logger to a STAT message

    Parameters
    ----------
    None

    Returns
    -------
    message : str
        The message that was logged/printed

    Raises
    ------
    ValueError
        If global log is not set

    Notes
    -----
    This function is thread-safe
    """
    if global_log is None:
        raise ValueError("This function requires the global log to be set")
    return global_log.log_profilers()



from threading import Lock
from logging import Logger
from types import FrameType
from typing import Any
import logging
import inspect
import os
import shutil
import sys
import getpass
import traceback
import time
import platform
import atexit
import weakref
from enum import IntEnum
from datetime import datetime, timedelta

from .utils import add_logging_level
from .screenshots import ScreenshotManager
from .global_logger import set_global_log
from .protocols import IProfiler


class LogLevel(IntEnum):
    DEBUG = 1
    INFO = 2
    STAT = 3
    WARNING = 4
    ERROR = 5
    EXCEPTION = 6
    CRITICAL = 7

    def to_logging_level(self):
        """Returns the corresponding logging level from the logging module."""
        if self == LogLevel.DEBUG:
            return logging.DEBUG
        elif self == LogLevel.INFO:
            return logging.INFO
        elif self == LogLevel.STAT:
            return logging.INFO + 5
        elif self == LogLevel.WARNING:
            return logging.WARNING
        elif self == LogLevel.ERROR:
            return logging.ERROR
        elif self == LogLevel.EXCEPTION:
            return logging.ERROR  # logging.exception also uses logging.ERROR level
        elif self == LogLevel.CRITICAL:
            return logging.CRITICAL
        else:
            raise ValueError(f"Unknown LogLevel: {self}")


class AutomationLogger:
    """
    Class used to log the program execution to a file

    Examples
    --------
    >>> log_dir = "./tests/Logs"
    >>> log = AutomationLogger(script_path=__file__, log_dir=log_dir, log_to_console=True)
    >>>
    >>> log.info("Start of program")
    >>> for i in range(10):
    >>>     log.info(f"Iteration index: {i}")
    >>> try:
    >>>     error = 1 / 0
    >>> except Exception as exc:
    >>>     log.exception(f"Exception caught: {repr(exc)}")
    >>> log.info("End of program")
    """

    def __init__(
        self,
        script_path: str,
        log_dir: str | None = None,
        log_to_console: bool = True,
        log_to_file: bool = True,
        max_logs: int = 60,
        max_days: int = 30,
        encoding: str = "utf-8",
        as_logging_root: bool = False,
        as_global_log: bool = True,
        level_threshold: LogLevel = LogLevel.INFO,
    ) -> None:
        """
        Initializes a AutomationLogger object for standardized logging in a script or robot.

        This class introduces a log level called "STAT" (short for statistics), designed
        for dumping dictionaries of information usable with logging solutions (e.g., ELK stack).

        Parameters
        ----------
        script_path : str
            Absolute path of the script calling this function. Use __file__ or
            os.path.abspath(__file__) to retrieve the path.

        log_dir : str, optional
            Path to the directory where logs will be stored. If None, logs will be stored
            in the directory script_path/Logs.

        log_to_console : bool, optional
            If True, log messages will also appear on the console.

        log_to_file : bool, optional
            If True, creates a log file and writes to it

        max_logs : int, optional
            Number of log files to retain in the log directory.

        max_days : int, optional
            Number of days of log history to retain in the log folder.

        encoding : str, optional
            Encoding used by the logger object.

        as_logging_root: bool, optional
            Flag that controls if the root logger of the `logging` library will use the same configurations as AutomationLogger
            When set, code that uses `logging` functions (e.g info(), error()...) will follow the AutomationLogger configuration

            One reason to keep it as False is to avoid polluting the log with messages from packages that use `logging` internally

        as_global_log: bool, optional
            Flag that controls if this logger object will be set as the module's global logger.
            When set, it will be possible to use the module's logging functions without using set_global_log

        level_threshold: LogLevel
            Lowest log level that will be logged, Any level less severe than the current one will be ignored.
            The order from least severe to most severe: debug, info, state, warning, error, exception, critical

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If `mode` passed is invalid
        ValueError
            If `max_logs` is not a positive integer
        ValueError
            If `max_days` is not a positive integer

        Examples
        --------
        >>> import logging
        >>> from automation_logging import AutomationLogger
        >>>
        >>> log_dir = "./tests/Logs"
        >>> logger = AutomationLogger(script_path=__file__, log_dir=log_dir, log_to_console=True, set_logging_root = True)
        >>> logger.info("This INFO message was written using a AutomationLogger method")
        >>> logging.info("This INFO message was written using a logging function and was written to the same file")
        """

        if not isinstance(max_logs, int) or max_logs <= 0:
            raise ValueError("max_logs parameter must be a positive integer")
        if not isinstance(max_days, int) or max_days <= 0:
            raise ValueError("max_days parameter must be a positive integer")
        if not isinstance(level_threshold, LogLevel):
            raise ValueError("level_threshold must be a LogLevel")
        if log_to_console is False and log_to_file is False:
            raise ValueError("log_to_console and log_to_file cannot be both False")

        log_dir = log_dir or os.path.join(os.path.dirname(script_path), "Logs")
        script_name = os.path.basename(script_path)
        script_name_no_ext = os.path.splitext(script_name)[0]
        log_name = f"{script_name_no_ext}.log"

        self.log_dir: str = log_dir
        self.log_name: str = log_name
        self.threshold: LogLevel = level_threshold
        # Include private mutex to make this class thread-safe
        self._mutex: Lock = Lock()

        # Initialize screenshot manager
        self._screenshot_manager: ScreenshotManager | None = None

        # Initialize an dict for storing profilers
        self._profilers: dict[str, IProfiler] = {}
        self.start_time: float = time.time()

        try:
            # Delete older files and create current directory
            if log_to_file:
                self.log_dir = self._clear_log_dir(
                    log_dir=log_dir,
                    max_logs=max_logs,
                    max_days=max_days,
                )
                self.log_file: str = os.path.join(self.log_dir, log_name)
            else:
                self.log_file = ""

            # Getting system and user information to include in the log
            system_name = platform.node()
            user_domain = os.environ.get("USERDOMAIN", "-")
            if user_domain == "-":
                complete_user = getpass.getuser()
            else:
                complete_user = f"{user_domain}\\{getpass.getuser()}"
            # Configuring the logger object
            self._logger: Logger = logging.getLogger(name=script_name)
            self._logger.setLevel(level_threshold.to_logging_level())

            self._logger.handlers = []  # Reset handlers

            # log to file
            if log_to_file:
                log_format = (
                    f"%(asctime)s | {system_name} | {complete_user} | %(name)s"
                    " | %(levelname)s | %(frame_info)s | %(message)s"
                )
                formatter = logging.Formatter(log_format, datefmt="%Y-%m-%d %H:%M:%S%z")
                file_handler = logging.FileHandler(self.log_file, encoding=encoding)
                file_handler.setLevel(level_threshold.to_logging_level())
                file_handler.setFormatter(formatter)
                self._logger.addHandler(file_handler)

            # print in console
            if log_to_console:
                log_format = "%(asctime)s | %(name)s | %(levelname)s | %(message)s"
                formatter = logging.Formatter(log_format, datefmt="%Y-%m-%d %H:%M:%S%z")
                console_handler = logging.StreamHandler(sys.stdout)
                console_handler.setLevel(level_threshold.to_logging_level())
                console_handler.setFormatter(formatter)
                self._logger.addHandler(console_handler)

            # Create new log level
            add_logging_level("STAT", LogLevel.STAT.to_logging_level())

            if as_logging_root:
                logging.root = self._logger
                logging.getLogger().handlers = self._logger.handlers
                logging.getLogger().setLevel(self._logger.level)

            if as_global_log:
                set_global_log(self)  # pyright: ignore[reportArgumentType]

            # Initialize screenshot manager after log directory is set
            if log_to_file:
                self._screenshot_manager = ScreenshotManager(self.log_dir, self._logger)

            _ = atexit.register(self._exit_weak())
            self.info("AutomationLogger object instantiated")
        except Exception as exc:
            print(f"Error when instantiating AutomationLogger object: {repr(exc)}")
            error_traceback = traceback.format_exc()
            print(error_traceback)
            print("Writing the error inside log_error.txt")
            file_path = os.path.abspath("log_error.txt")
            with open(file_path, "w+", encoding="utf-8") as file:
                _ = file.write(f"Error when instantiating AutomationLogger object: {repr(exc)}\n")
                _ = file.write("----------------------------------------------------------\n")
                _ = file.write(error_traceback)
            raise exc

    def _clear_log_dir(self, log_dir: str, max_logs: int, max_days: int) -> str:
        """
        Deletes old logs and creates the directory for the current log.

        Parameters
        ----------
        log_dir : str
            Path to the directory where logs will be stored. If None, logs will be stored
            in the directory specified during initialization.

        max_logs : int
            Number of log files to retain in the log directory.

        max_days : int
            Number of days of log history to retain in the log folder.

        Returns
        -------
        str
            String with the folder path for the current execution.
        """

        os.makedirs(log_dir, exist_ok=True)

        timestamp = datetime.now().strftime("exec_%Y-%m-%d_%H-%M-%S")
        # Delete older execution if limit of executions has been reached
        # Condition 1) Delete based on the number of log files
        files = sorted(
            [
                x
                for x in os.listdir(log_dir)
                if os.path.isdir(os.path.join(log_dir, x)) and "exec_" in x.lower()
            ]
        )
        num_delete = max(len(files) - max_logs + 1, 0)
        for i in range(num_delete):
            try:
                shutil.rmtree(os.path.join(log_dir, files[i]))
            except PermissionError as exc:
                # In case of permission error when deleting files, instead of crashing the program,
                # the files won't be deleted (need to be manually deleted)
                print(f"Error when deleting old log files: {repr(exc)}")

        # Condition 2) Delete based on date in the file name and a cutoff date
        cutoff_date = datetime.now() - timedelta(days=max_days)
        files = [
            x
            for x in os.listdir(log_dir)
            if os.path.isdir(os.path.join(log_dir, x)) and "exec_" in x.lower()
        ]
        for file in files:
            try:
                file_date = datetime.strptime(file, "exec_%Y-%m-%d_%H-%M-%S")
                if file_date <= cutoff_date:
                    shutil.rmtree(os.path.join(log_dir, file))
            except Exception as exc:
                print(f"Error when deleting old log files: {repr(exc)}")

        log_dir = os.path.join(log_dir, timestamp)

        # Create log dir for the current execution
        os.makedirs(log_dir, exist_ok=True)
        return log_dir

    def _get_frame_info(self) -> str:
        """Get frame info from the last frame before entering automation_logging modules"""

        nf_alog = False
        cf_alog = False
        frames: list[tuple[FrameType, str]] = []
        frame = inspect.currentframe()
        while frame:
            frames.append((frame, inspect.getfile(frame)))
            frame = frame.f_back
        frame_data: dict[str, Any] = {}
        for i in range(len(frames) - 1, 0, -1):
            cf = frames[i]
            nf = frames[i - 1]
            cf_alog = "automation_logging" in cf[1]
            nf_alog = "automation_logging" in nf[1]
            if nf_alog and not cf_alog:
                frame = cf[0]
                frame_data["module"] = inspect.getmodulename(inspect.getfile(frame))
                frame_data["lineno"] = inspect.getlineno(frame)
                frame_data["function"] = frame.f_code.co_name
                break

        if len(frame_data) == 0:
            return "-"
        return f"{frame_data['module']}.{frame_data['function']}:{frame_data['lineno']}"

    def _write(self, message: str, level: LogLevel) -> None:
        """Writes the message to the log file with the given level
        If the level is lower than the threshold, the message is not written

        Parameters
        ----------
        message: str
            Message to be logged

        Returns
        -------
        None
        """

        if level < self.threshold:
            return None

        frame_info = self._get_frame_info()

        with self._mutex:
            if level == LogLevel.DEBUG:
                self._logger.debug(message, extra={"frame_info": frame_info})
            elif level == LogLevel.INFO:
                self._logger.info(message, extra={"frame_info": frame_info})
            elif level == LogLevel.STAT:
                self._logger.stat(message, extra={"frame_info": frame_info})  # pyright: ignore[reportAttributeAccessIssue, reportUnknownMemberType]
            elif level == LogLevel.WARNING:
                self._logger.warning(message, extra={"frame_info": frame_info})
            elif level == LogLevel.ERROR:
                self._logger.error(message, extra={"frame_info": frame_info})
            elif level == LogLevel.EXCEPTION:
                self._logger.exception(message, extra={"frame_info": frame_info})
            elif level == LogLevel.CRITICAL:
                self._logger.critical(message, extra={"frame_info": frame_info})

    def debug(self, message: str) -> None:
        """Writes the message to the log file with level DEBUG

        Parameters
        ----------
        message: str
            Message to be logged

        Returns
        -------
        None

        Examples
        --------
        >>> log = AutomationLogger(script_path=__file__)
        >>> log.debug(f"This is a DEBUG example")
        """

        self._write(message, LogLevel.DEBUG)

    def info(self, message: str) -> None:
        """
        Writes the message to the log file with level INFO.

        Parameters
        ----------
        message: str
            Message to be logged

        Returns
        -------
        None

        Examples
        --------
        >>> log = AutomationLogger(script_path=__file__)
        >>> log.info(f"This is an INFO example")
        """

        self._write(message, LogLevel.INFO)

    def warning(self, message: str) -> None:
        """
        Writes the message to the log file with level WARNING.

        Parameters
        ----------
        message: str
            Message to be logged

        Returns
        -------
        None

        Examples
        --------
        >>> log = AutomationLogger(script_path=__file__)
        >>> log.warning("This is a WARNING example")
        """

        self._write(message, LogLevel.WARNING)

    def error(self, message: str) -> None:
        """
        Writes the message to the log file with level ERROR.

        The difference between log.error and log.exception is that log.exception also prints
        the stack traceback of the exception.

        Parameters
        ----------
        message: str
            Message to be logged

        Returns
        -------
        None

        Examples
        --------
        >>> log = AutomationLogger(script_path=__file__)
        >>> try:
        >>>     a = 1/0
        >>> except Exception as exc:
        >>>     log.error(f"This is an ERROR example: {repr(exc)}")
        """

        self._write(message, LogLevel.ERROR)

    def critical(self, message: str) -> None:
        """
        Writes the message to the log file with level CRITICAL.

        Parameters
        ----------
        message: str
            Message to be logged

        Returns
        -------
        None

        Examples
        --------
        >>> log = AutomationLogger(script_path=__file__)
        >>> is_valid = check_critical_component(...)
        >>> if not is_valid:
        >>>     log.critical("This is a CRITICAL example, cannot continue")
        >>>     sys.exit(2)
        """

        self._write(message, LogLevel.CRITICAL)

    def exception(self, message: str) -> None:
        """
        Writes an exception to the log file with level ERROR.

        The difference between log.error and log.exception is that log.exception also prints
        the stack traceback of the exception.

        Parameters
        ----------
        message: str
            Message to be logged

        Returns
        -------
        None

        Examples
        --------
        >>> log = AutomationLogger(script_path=__file__)
        >>> try:
        >>>     a = 1/0
        >>> except Exception as exc:
        >>>     log.exception(f"This is an EXCEPTION example: {repr(exc)}")
        """

        self._write(message, LogLevel.EXCEPTION)

    def stat(self, message: str | dict[Any, Any]) -> None:
        """
        Writes a dictionary to the log file with level STAT.

        This method should be used to aggregate or summarize information in a single message
        to facilitate the use of analytics solutions like elastic and opensearch.

        Parameters
        ----------
        message: str | dict[Any, Any]
            Message to be logged

        Returns
        -------
        None

        Examples
        --------
        >>> log = AutomationLogger(script_path=__file__)
        >>> info_dict = {"program_name": "Test Program", "var1": "StringVariable", "var2": 3}
        >>> log.stat(info_dict)
        """

        self._write(str(message), LogLevel.STAT)

    def capture_screenshot(self, filename: str | None = None, optimize_size: bool = False) -> str:
        """
        Captures a screenshot of the entire screen and saves it in the log directory.

        Parameters
        ----------
        filename : str, optional
            Name of the screenshot.
        optimize_size : bool
            Flag to save the image using a memory efficient and lossless format (.webp)

        Returns
        -------
        filename : str
            The name of the file saved

        Raises
        ------
        ValueError
            If screenshot manager is not available (log_to_file=False)
        """
        if self._screenshot_manager is None:
            raise ValueError("Screenshot functionality requires log_to_file=True")
        return self._screenshot_manager.capture_screenshot(filename, optimize_size)

    def capture_screenshot_selenium(self, driver: Any, filename: str | None = None) -> str:
        """
        Captures a screenshot of a Selenium driver instance.

        Parameters
        ----------
        driver : WebDriver
            Selenium WebDriver instance from which to capture the screenshot.
        filename : str, optional
            Name of the screenshot.

        Returns
        -------
        filename : str
            The name of the file saved

        Raises
        ------
        ValueError
            If screenshot manager is not available (log_to_file=False)
        """
        if self._screenshot_manager is None:
            raise ValueError("Screenshot functionality requires log_to_file=True")
        return self._screenshot_manager.capture_screenshot_selenium(driver, filename)

    def group_by_prefix(self, prefix: str | None = None, sep: str | None = None):
        """
        Group files in the log directory by prefix.

        Parameters
        ----------
        prefix: str, optional
            Prefix of files that will be grouped
        sep: str, optional
            Separator used to split file names and automatically determine file prefixes.

        Raises
        ------
        ValueError
            If screenshot manager is not available (log_to_file=False)
        """
        if self._screenshot_manager is None:
            raise ValueError("File grouping functionality requires log_to_file=True")
        return self._screenshot_manager.group_by_prefix(prefix, sep)

    def insert_profiler(self, name: str, prof: IProfiler):
        """Insert profiler if not in the dict"""

        if name not in self._profilers:
            self._profilers[name] = prof

    def _exit_weak(self):
        weak_self = weakref.ref(self)

        def callback():
            obj = weak_self()
            if obj is not None:
                return obj.log_profilers()

        return callback

    def log_profilers(self) -> str:
        """
        Write profilers captured by the logger to a STAT message

        Parameters
        ----------
        None

        Returns
        -------
        message : str
            The message that was logged/printed

        Raises
        ------
        None
        """

        num_profilers = len(self._profilers)
        if num_profilers == 0:
            return ""

        elapsed_time_logger = time.time() - self.start_time
        message = (
            f"Logger instance holds {num_profilers} {'profiler' if num_profilers == 1 else 'profilers'}. "
            f"Time since instance was created: {elapsed_time_logger} seconds.\n"
        )
        for prof in self._profilers.values():
            message += "> " + repr(prof) + "\n"
        self.stat(message)
        return message



import os
from importlib.util import find_spec

# Environment variable configuration
truthy_values = ["1", "true", "t", "yes", "y", "on"]
OPT_DISABLE_WEB = os.environ.get("alog_disable_web", "0").lower() in truthy_values
OPT_DISABLE_IMAGE = os.environ.get("alog_disable_image", "0").lower() in truthy_values

# Selenium dependency check
if OPT_DISABLE_WEB:
    web_enabled = False
else:
    web_enabled = find_spec("selenium") is not None

# Image capture dependency check
if OPT_DISABLE_IMAGE:
    image_enabled = False
else:
    image_enabled = find_spec(name="pyautogui") is not None and find_spec(name="PIL") is not None

# Import screenshot function if available
if image_enabled:
    from pyautogui import screenshot  # pyright: ignore[reportAssignmentType]
    from PIL.Image import Image  # pyright: ignore[reportAssignmentType]
else:

    class Image:
        """Dummy image class"""

        def save(self, *args, **kwargs) -> None:
            return None

    # Dummy implementations
    def screenshot(*args, **kwargs) -> Image:
        return Image()



import logging


def add_logging_level(levelName: str, levelNum: int, methodName: str | None = None) -> None:
    """
    Comprehensively adds a new logging level to the `logging` module and the
    currently configured logging class.

    `levelName` becomes an attribute of the `logging` module with the value
    `levelNum`. `methodName` becomes a convenience method for both `logging`
    itself and the class returned by `logging.getLoggerClass()` (usually just
    `logging.Logger`). If `methodName` is not specified, `levelName.lower()` is
    used.

    Parameters
    ----------
    levelName : str
        Name of the new logging level.

    levelNum : int
        Numeric value assigned to the new logging level.

    methodName : str, optional
        Name of the convenience method for the new logging level. If not specified,
        `levelName.lower()` is used.

    Returns
    -------
    None

    Examples
    --------
    >>> add_logging_level('TRACE', logging.DEBUG - 5)
    >>> logging.getLogger(__name__).setLevel("TRACE")
    >>> logging.getLogger(__name__).trace('that worked')
    >>> logging.trace('so did this')
    >>> logging.TRACE
    5

    Notes
    -----
    This method was extracted from:
    https://stackoverflow.com/q/2183233/35804945#35804945
    The behavior was modified to return None if the level is already defined
    instead of throwing an exception, allowing re-initialization of the log.
    """
    if not methodName:
        methodName = levelName.lower()

    if (
        hasattr(logging, levelName)
        or hasattr(logging, methodName)
        or hasattr(logging.getLoggerClass(), methodName)
    ):
        return None

    def logForLevel(self, message, *args, **kwargs):
        if self.isEnabledFor(levelNum):
            self._log(levelNum, message, args, **kwargs)

    def logToRoot(message, *args, **kwargs):
        logging.log(levelNum, message, *args, **kwargs)

    logging.addLevelName(levelNum, levelName)
    setattr(logging, levelName, levelNum)
    setattr(logging.getLoggerClass(), methodName, logForLevel)
    setattr(logging, methodName, logToRoot)



from typing import Any, Protocol, runtime_checkable


@runtime_checkable
class IProfiler(Protocol):
    pass


@runtime_checkable
class IAutomationLogger(Protocol):
    def debug(self, message: str) -> None: ...
    def info(self, message: str) -> None: ...
    def stat(self, message: str | dict[Any, Any]) -> None: ...
    def warning(self, message: str) -> None: ...
    def error(self, message: str) -> None: ...
    def exception(self, message: str) -> None: ...
    def critical(self, message: str) -> None: ...
    def capture_screenshot(
        self, filename: str | None = None, optimize_size: bool = False
    ) -> str: ...
    def capture_screenshot_selenium(self, driver: Any, filename: str | None = None) -> str: ...
    def group_by_prefix(self, prefix: str | None = None, sep: str | None = None) -> None: ...
    def insert_profiler(self, name: str, prof: IProfiler) -> None: ...
    def log_profilers(self) -> str: ...



from protocols import IAutomationLogger


from profiler import Profiler
import global_logger as log
import core

if __name__ == "__main__":
    import time

    core.AutomationLogger("test", "local_logs")

    @Profiler("ms")
    def foo():
        time.sleep(0.1)

    t1 = time.monotonic()
    for _ in range(2):
        foo()
    t2 = time.monotonic()
    print(t2 - t1)
    print(foo.__getattribute__("profiler"))

    logger: IAutomationLogger | None = log.get_global_log()
    assert isinstance(logger, core.AutomationLogger), "error"
    pass



"""
Automation Logging
====
- Thread-Safe Logging: Ensures reliable and consistent logging from multi-threaded applications, preventing message corruption.
- Automated Log Management: Simplifies log file maintenance with configurable retention policies based on age or file count.
- Seamless Integration: Enables effortless integration with existing projects by allowing the library to override the standard `logging` module's root logger.
- Integrated Screenshot Capture: Captures informative screenshots directly within your logs, supporting full-screen captures (using `pyautogui`) and Selenium WebDriver screenshots, including headless browser support (using `selenium`).
- Performance Profiling: Capture elapsed and CPU times of functions using decorators and log the results.
"""

__version__ = "0.1.0"

# Import core components
from .core import LogLevel, AutomationLogger
from .screenshots import ScreenshotManager
from .utils import add_logging_level
from .global_logger import (
    set_global_log,
    get_global_log,
    debug,
    debug_else,
    info,
    info_else,
    stat,
    stat_else,
    warning,
    warning_else,
    error,
    error_else,
    exception,
    exception_else,
    critical,
    critical_else,
    capture_screenshot,
    capture_screenshot_selenium,
    group_by_prefix,
    log_profilers,
)
from .config import image_enabled, web_enabled
from .profiler import Profiler

__all__ = [
    # Core classes
    "LogLevel",
    "AutomationLogger",
    "ScreenshotManager",
    "Profiler",
    # Global logging functions
    "set_global_log",
    "get_global_log",
    "debug",
    "debug_else",
    "info",
    "info_else",
    "stat",
    "stat_else",
    "warning",
    "warning_else",
    "error",
    "error_else",
    "exception",
    "exception_else",
    "critical",
    "critical_else",
    "capture_screenshot",
    "capture_screenshot_selenium",
    "group_by_prefix",
    "log_profilers",
    # Utilities
    "add_logging_level",
    # Configuration
    "web_enabled",
    "image_enabled",
]
